import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset
import ray
import numpy as np

# Initialize Ray
ray.init()

# Define a simple PyTorch model
class SimpleModel(nn.Module):
    def __init__(self):
        super(SimpleModel, self).__init__()
        self.fc1 = nn.Linear(10, 50)
        self.fc2 = nn.Linear(50, 1)
    
    def forward(self, x):
        x = torch.relu(self.fc1(x))
        x = self.fc2(x)
        return x

# Remote function to perform training on each worker
@ray.remote
def train_worker(rank, num_epochs=5):
    # Create random data for training (100 samples, 10 features)
    X = torch.randn(100, 10)
    y = torch.randn(100, 1)
    
    # Create DataLoader
    dataset = TensorDataset(X, y)
    dataloader = DataLoader(dataset, batch_size=16, shuffle=True)

    # Initialize model, loss function, and optimizer
    model = SimpleModel()
    criterion = nn.MSELoss()
    optimizer = optim.SGD(model.parameters(), lr=0.01)

    # Training loop
    for epoch in range(num_epochs):
        running_loss = 0.0
        for batch_X, batch_y in dataloader:
            optimizer.zero_grad()
            outputs = model(batch_X)
            loss = criterion(outputs, batch_y)
            loss.backward()
            optimizer.step()
            running_loss += loss.item()
        
        print(f"Worker {rank} - Epoch {epoch + 1}/{num_epochs}, Loss: {running_loss / len(dataloader)}")

    # Return the trained model parameters
    return model.state_dict()

# Aggregating the results (optional)
def average_model_params(models):
    avg_state_dict = models[0]
    for key in avg_state_dict.keys():
        for i in range(1, len(models)):
            avg_state_dict[key] += models[i][key]
        avg_state_dict[key] = avg_state_dict[key] / len(models)
    return avg_state_dict

if __name__ == "__main__":
    # Number of workers
    num_workers = 4

    # Launch distributed training on multiple workers
    futures = [train_worker.remote(rank=i, num_epochs=5) for i in range(num_workers)]

    # Collect the trained model parameters from each worker
    results = ray.get(futures)

    # Aggregate the model parameters from all workers (averaging)
    avg_model_params = average_model_params(results)

    print("Aggregated model parameters from all workers!")

    # Shutdown Ray
    ray.shutdown()